"""FastAPI routes for MCP Resource Server authorization endpoints.

The routes defined here enable MCP clients to discover authorization servers
associated with this MCP server.
"""

import logging
from urllib.parse import urlparse

from fastapi import APIRouter
from fastapi.responses import JSONResponse

from arcade_mcp_server.resource_server.base import ResourceServerValidator

logger = logging.getLogger(__name__)


def create_auth_router(
    resource_server_validator: ResourceServerValidator,
    canonical_url: str | None,
) -> APIRouter:
    """Create FastAPI router with OAuth discovery endpoints.

    The well-known URI is constructed by inserting the well-known path after the host.
    If the canonical URL has a path component, then it becomes a suffix on the well-known path.

    For example:
    - canonical_url "https://example.com" -> "/.well-known/oauth-protected-resource"
    - canonical_url "https://example.com/mcp" -> "/.well-known/oauth-protected-resource/mcp"

    Args:
        resource_server_validator: The resource server validator instance
        canonical_url: Canonical URL of the MCP server

    Returns:
        APIRouter configured with OAuth discovery endpoints
    """
    router = APIRouter(tags=["MCP Protocol"])

    path_suffix = ""
    if canonical_url:
        parsed = urlparse(canonical_url)
        path_suffix = parsed.path

    well_known_base = "/.well-known/oauth-protected-resource"
    well_known_path = f"{well_known_base}{path_suffix}"

    async def oauth_protected_resource() -> JSONResponse:
        """OAuth 2.0 Protected Resource Metadata (RFC 9728)"""
        if not canonical_url:
            return JSONResponse(
                {"error": "Server canonical URL not configured"},
                status_code=500,
            )

        metadata = resource_server_validator.get_resource_metadata()
        if metadata is None:
            logger.error(
                "Resource metadata unavailable for OAuth discovery endpoint. "
                "This is unexpected - the validator should provide metadata if OAuth discovery is enabled."
            )
            return JSONResponse(
                {"error": "Resource metadata not available"},
                status_code=500,
                headers={
                    "Access-Control-Allow-Origin": "*",
                    "Access-Control-Allow-Methods": "GET, OPTIONS",
                    "Access-Control-Allow-Headers": "Content-Type",
                },
            )

        return JSONResponse(
            metadata,
            headers={
                "Access-Control-Allow-Origin": "*",
                "Access-Control-Allow-Methods": "GET, OPTIONS",
                "Access-Control-Allow-Headers": "Content-Type",
            },
        )

    # Register the well-known endpoint at the RFC 9728 compliant path
    router.add_api_route(
        well_known_path,
        oauth_protected_resource,
        methods=["GET"],
        name="oauth_protected_resource",
    )

    # Also register at base path if there's a suffix for extra compatibility
    if path_suffix:
        router.add_api_route(
            well_known_base,
            oauth_protected_resource,
            methods=["GET"],
            include_in_schema=False,
        )

    return router



"""
MCP Exception Hierarchy

Provides domain-specific exceptions for better error handling and debugging.
"""

from arcade_core.errors import (
    ContextRequiredToolError,
    ErrorKind,
    FatalToolError,
    RetryableToolError,
    ToolExecutionError,
    ToolRuntimeError,
    UpstreamError,
    UpstreamRateLimitError,
)

__all__ = [
    # Re-exports
    "ErrorKind",
    "FatalToolError",
    "RetryableToolError",
    "ToolExecutionError",
    "ToolRuntimeError",
    "UpstreamError",
    "UpstreamRateLimitError",
    "ContextRequiredToolError",
    # Base exceptions
    "MCPError",
    "MCPRuntimeError",
    # Server exceptions
    "ServerError",
    "SessionError",
    "RequestError",
    "ResponseError",
    "ServerRequestError",
    "LifespanError",
    # Context exceptions
    "MCPContextError",
    "NotFoundError",
    "AuthorizationError",
    "PromptError",
    "ResourceError",
    "TransportError",
    "ProtocolError",
]


class MCPError(Exception):
    """Base error for all MCP-related exceptions."""


class MCPRuntimeError(MCPError):
    """Runtime error for all MCP-related exceptions."""


class ServerError(MCPRuntimeError):
    """Error in server operations."""


class SessionError(ServerError):
    """Error in session management"""


class RequestError(ServerError):
    """Error in request processing from client to server"""


class ResponseError(ServerError):
    """Error in request processing from server -> client"""


class ServerRequestError(RequestError):
    """Error in sending request from server -> client initiated by the server"""


class LifespanError(ServerError):
    """Error in lifespan management."""


class MCPContextError(MCPError):
    """Error in context management."""


class NotFoundError(MCPContextError):
    """Requested entity not found."""


class AuthorizationError(MCPContextError):
    """Authorization failure."""


class PromptError(MCPContextError):
    """Error in prompt management."""


class ResourceError(MCPContextError):
    """Error in resource management."""


# Transport and Protocol Errors


class TransportError(MCPRuntimeError):
    """Error in transport layer (stdio, HTTP, etc)."""


class ProtocolError(MCPRuntimeError):
    """Error in MCP protocol handling."""



from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum
from typing import Any, Generic, Literal, TypeAlias, TypeVar

from pydantic import BaseModel, ConfigDict, Field

from arcade_mcp_server.resource_server.base import ResourceOwner

# -----------------------------------------------------------------------------
# JSON-RPC constants
# -----------------------------------------------------------------------------

JSONRPC_VERSION: Literal["2.0"] = "2.0"
LATEST_PROTOCOL_VERSION: str = "2025-06-18"

# -----------------------------------------------------------------------------
# Basic types
# -----------------------------------------------------------------------------

ProgressToken = str | int
Cursor = str
RequestId = str | int
AnyFunction: TypeAlias = Callable[..., Any]


# -----------------------------------------------------------------------------
# Base JSON-RPC shapes
# -----------------------------------------------------------------------------


class Request(BaseModel):
    method: str
    params: Any = None

    model_config = ConfigDict(extra="allow", populate_by_name=True)


class Notification(BaseModel):
    method: str
    params: Any = None

    model_config = ConfigDict(extra="allow", populate_by_name=True)


class Result(BaseModel):
    meta: dict[str, Any] | None = Field(alias="_meta", default=None)

    model_config = ConfigDict(extra="allow", populate_by_name=True)


class JSONRPCMessage(BaseModel):
    jsonrpc: Literal["2.0"] = Field(default=JSONRPC_VERSION, frozen=True)

    model_config = ConfigDict(extra="allow")


class JSONRPCRequest(JSONRPCMessage, Request):
    id: RequestId


T = TypeVar("T", bound=Result)


class JSONRPCResponse(JSONRPCMessage, Generic[T]):
    id: RequestId
    result: T | dict[str, Any]


# Standard JSON-RPC error codes
PARSE_ERROR = -32700
INVALID_REQUEST = -32600
METHOD_NOT_FOUND = -32601
INVALID_PARAMS = -32602
INTERNAL_ERROR = -32603


class ErrorData(BaseModel):
    code: int
    message: str
    data: Any | None = None


class JSONRPCError(JSONRPCMessage):
    id: RequestId
    error: dict[str, Any]


# -----------------------------------------------------------------------------
# Transport types
# -----------------------------------------------------------------------------


@dataclass
class SessionMessage:
    """Wrapper for messages in transport sessions.

    Carries both the MCP protocol message and optional authenticated user
    information from front-door authentication.
    """

    message: JSONRPCMessage
    resource_owner: ResourceOwner | None = None


# -----------------------------------------------------------------------------
# Initialization
# -----------------------------------------------------------------------------


class BaseMetadata(BaseModel):
    name: str
    title: str | None = None

    model_config = ConfigDict(extra="allow")


class Implementation(BaseMetadata):
    version: str


class ClientCapabilities(BaseModel):
    experimental: dict[str, object] | None = None
    roots: dict[str, Any] | None = None
    sampling: dict[str, Any] | None = None
    elicitation: dict[str, Any] | None = None

    model_config = ConfigDict(extra="allow")


class ServerCapabilities(BaseModel):
    experimental: dict[str, object] | None = None
    logging: dict[str, Any] | None = None
    completions: dict[str, Any] | None = None
    prompts: dict[str, Any] | None = None
    resources: dict[str, Any] | None = None
    tools: dict[str, Any] | None = None

    model_config = ConfigDict(extra="allow")


class InitializeParams(BaseModel):
    protocolVersion: str
    capabilities: ClientCapabilities = Field(default_factory=ClientCapabilities)
    clientInfo: Implementation


class InitializeRequest(JSONRPCRequest):
    method: Literal["initialize"] = Field(default="initialize", frozen=True)
    params: InitializeParams


class InitializeResult(Result):
    protocolVersion: str
    capabilities: ServerCapabilities
    serverInfo: Implementation
    instructions: str | None = None


class InitializedNotification(JSONRPCMessage, Notification):
    method: Literal["notifications/initialized"] = Field(
        default="notifications/initialized", frozen=True
    )


# -----------------------------------------------------------------------------
# Ping
# -----------------------------------------------------------------------------


class PingRequest(JSONRPCRequest):
    method: Literal["ping"] = Field(default="ping", frozen=True)


# -----------------------------------------------------------------------------
# Progress notifications
# -----------------------------------------------------------------------------


class ProgressNotificationParams(BaseModel):
    progressToken: ProgressToken
    progress: float
    total: float | None = None
    message: str | None = None


class ProgressNotification(JSONRPCMessage, Notification):
    method: Literal["notifications/progress"] = Field(default="notifications/progress", frozen=True)
    params: ProgressNotificationParams


# -----------------------------------------------------------------------------
# Pagination
# -----------------------------------------------------------------------------


class PaginatedRequest(JSONRPCRequest):
    params: dict[str, Any] | None = None


class PaginatedResult(Result):
    nextCursor: Cursor | None = None


# -----------------------------------------------------------------------------
# Annotations (used across resources, content, etc.)
# -----------------------------------------------------------------------------

Role = Literal["user", "assistant"]


class Annotations(BaseModel):
    audience: list[Role] | None = None
    priority: float | None = None
    lastModified: str | None = None

    model_config = ConfigDict(extra="allow")


# -----------------------------------------------------------------------------
# Resources
# -----------------------------------------------------------------------------


class Resource(BaseMetadata):
    uri: str
    description: str | None = None
    mimeType: str | None = None
    annotations: Annotations | None = None
    size: int | None = None
    meta: dict[str, Any] | None = Field(alias="_meta", default=None)


class ListResourcesRequest(PaginatedRequest):
    method: Literal["resources/list"] = Field(default="resources/list", frozen=True)


class ListResourcesResult(PaginatedResult):
    resources: list[Resource] = Field(default_factory=list)


class ListResourceTemplatesRequest(PaginatedRequest):
    method: Literal["resources/templates/list"] = Field(
        default="resources/templates/list", frozen=True
    )


class ResourceTemplate(BaseMetadata):
    uriTemplate: str
    description: str | None = None
    mimeType: str | None = None
    annotations: Annotations | None = None
    meta: dict[str, Any] | None = Field(alias="_meta", default=None)


class ListResourceTemplatesResult(PaginatedResult):
    resourceTemplates: list[ResourceTemplate] = Field(default_factory=list)


class ReadResourceParams(BaseModel):
    uri: str


class ReadResourceRequest(JSONRPCRequest):
    method: Literal["resources/read"] = Field(default="resources/read", frozen=True)
    params: ReadResourceParams


class ResourceContents(BaseModel):
    uri: str
    mimeType: str | None = None
    meta: dict[str, Any] | None = Field(alias="_meta", default=None)


class TextResourceContents(ResourceContents):
    text: str


class BlobResourceContents(ResourceContents):
    blob: str


class ReadResourceResult(Result):
    contents: list[TextResourceContents | BlobResourceContents]


class ResourceListChangedNotification(JSONRPCMessage, Notification):
    method: Literal["notifications/resources/list_changed"] = Field(
        default="notifications/resources/list_changed", frozen=True
    )


class ResourceUpdatedNotificationParams(BaseModel):
    uri: str


class ResourceUpdatedNotification(JSONRPCMessage, Notification):
    method: Literal["notifications/resources/updated"] = Field(
        default="notifications/resources/updated", frozen=True
    )
    params: ResourceUpdatedNotificationParams


class SubscribeParams(BaseModel):
    uri: str


class SubscribeRequest(JSONRPCRequest):
    method: Literal["resources/subscribe"] = Field(default="resources/subscribe", frozen=True)
    params: SubscribeParams


class UnsubscribeParams(BaseModel):
    uri: str


class UnsubscribeRequest(JSONRPCRequest):
    method: Literal["resources/unsubscribe"] = Field(default="resources/unsubscribe", frozen=True)
    params: UnsubscribeParams


# -----------------------------------------------------------------------------
# Prompts
# -----------------------------------------------------------------------------


class PromptArgument(BaseMetadata):
    description: str | None = None
    required: bool | None = None


class Prompt(BaseMetadata):
    description: str | None = None
    arguments: list[PromptArgument] | None = None
    meta: dict[str, Any] | None = Field(alias="_meta", default=None)


class ListPromptsRequest(PaginatedRequest):
    method: Literal["prompts/list"] = Field(default="prompts/list", frozen=True)


class ListPromptsResult(PaginatedResult):
    prompts: list[Prompt] = Field(default_factory=list)


class PromptMessage(BaseModel):
    role: Role
    content: dict[str, Any]


class GetPromptParams(BaseModel):
    name: str
    arguments: dict[str, str] | None = None


class GetPromptRequest(JSONRPCRequest):
    method: Literal["prompts/get"] = Field(default="prompts/get", frozen=True)
    params: GetPromptParams


class GetPromptResult(Result):
    description: str | None = None
    messages: list[PromptMessage]


class PromptListChangedNotification(JSONRPCMessage, Notification):
    method: Literal["notifications/prompts/list_changed"] = Field(
        default="notifications/prompts/list_changed", frozen=True
    )


# -----------------------------------------------------------------------------
# Tools
# -----------------------------------------------------------------------------


class ToolAnnotations(BaseModel):
    title: str | None = None
    readOnlyHint: bool | None = None
    destructiveHint: bool | None = None
    idempotentHint: bool | None = None
    openWorldHint: bool | None = None

    model_config = ConfigDict(extra="allow")


class MCPTool(BaseModel):
    name: str
    description: str | None = None
    inputSchema: dict[str, Any]
    outputSchema: dict[str, Any] | None = None
    annotations: ToolAnnotations | None = None
    title: str | None = None
    meta: dict[str, Any] | None = Field(alias="_meta", default=None)


class ListToolsRequest(PaginatedRequest):
    method: Literal["tools/list"] = Field(default="tools/list", frozen=True)


class ListToolsResult(PaginatedResult):
    tools: list[MCPTool]


class ToolListChangedNotification(JSONRPCMessage, Notification):
    method: Literal["notifications/tools/list_changed"] = Field(
        default="notifications/tools/list_changed", frozen=True
    )


class CallToolParams(BaseModel):
    name: str
    arguments: dict[str, Any] | None = None
    meta: dict[str, Any] | None = Field(alias="_meta", default=None)

    model_config = ConfigDict(extra="allow", populate_by_name=True)


class CallToolRequest(JSONRPCRequest):
    method: Literal["tools/call"] = Field(default="tools/call", frozen=True)
    params: CallToolParams


class TextContent(BaseModel):
    type: Literal["text"]
    text: str
    annotations: Annotations | None = None
    meta: dict[str, Any] | None = Field(alias="_meta", default=None)


class ImageContent(BaseModel):
    type: Literal["image"]
    data: str
    mimeType: str
    annotations: Annotations | None = None
    meta: dict[str, Any] | None = Field(alias="_meta", default=None)


class AudioContent(BaseModel):
    type: Literal["audio"]
    data: str
    mimeType: str
    annotations: Annotations | None = None
    meta: dict[str, Any] | None = Field(alias="_meta", default=None)


class ResourceLink(Resource):
    type: Literal["resource_link"] = Field(default="resource_link", frozen=True)


class EmbeddedResource(BaseModel):
    type: Literal["resource"] = Field(default="resource", frozen=True)
    resource: TextResourceContents | BlobResourceContents
    annotations: Annotations | None = None
    meta: dict[str, Any] | None = Field(alias="_meta", default=None)


MCPContent = TextContent | ImageContent | AudioContent | ResourceLink | EmbeddedResource


class CallToolResult(Result):
    """
    A list of content objects that represent the unstructured result of the tool call.
    """

    content: list[MCPContent]

    """
    An optional JSON object that represents the structured result of the tool call.
    """
    structuredContent: dict[str, Any] | None = None

    isError: bool | None = None


# -----------------------------------------------------------------------------
# Logging
# -----------------------------------------------------------------------------


class LoggingLevel(str, Enum):
    DEBUG = "debug"
    INFO = "info"
    NOTICE = "notice"
    WARNING = "warning"
    ERROR = "error"
    CRITICAL = "critical"
    ALERT = "alert"
    EMERGENCY = "emergency"


class SetLevelParams(BaseModel):
    level: LoggingLevel


class SetLevelRequest(JSONRPCRequest):
    method: Literal["logging/setLevel"] = Field(default="logging/setLevel", frozen=True)
    params: SetLevelParams


class LoggingMessageParams(BaseModel):
    level: LoggingLevel
    logger: str | None = None
    data: Any


class LoggingMessageNotification(JSONRPCMessage, Notification):
    method: Literal["notifications/message"] = Field(default="notifications/message", frozen=True)
    params: LoggingMessageParams


# -----------------------------------------------------------------------------
# Cancellation (notification-only)
# -----------------------------------------------------------------------------


class CancelledParams(BaseModel):
    requestId: RequestId
    reason: str | None = None


class CancelledNotification(JSONRPCMessage, Notification):
    method: Literal["notifications/cancelled"] = Field(
        default="notifications/cancelled", frozen=True
    )
    params: CancelledParams


# -----------------------------------------------------------------------------
# Sampling (server -> client)
# -----------------------------------------------------------------------------


class SamplingMessage(BaseModel):
    role: Role
    content: TextContent | ImageContent | AudioContent


class ModelHint(BaseModel):
    name: str | None = None


class ModelPreferences(BaseModel):
    hints: list[ModelHint] | None = None
    costPriority: float | None = None
    speedPriority: float | None = None
    intelligencePriority: float | None = None


class CreateMessageParams(BaseModel):
    messages: list[SamplingMessage]
    modelPreferences: ModelPreferences | None = None
    systemPrompt: str | None = None
    includeContext: Literal["none", "thisServer", "allServers"] | None = None
    temperature: float | None = None
    maxTokens: int
    stopSequences: list[str] | None = None
    metadata: dict[str, Any] | None = None


class CreateMessageRequest(JSONRPCRequest):
    method: Literal["sampling/createMessage"] = Field(default="sampling/createMessage", frozen=True)
    params: CreateMessageParams


class CreateMessageResult(Result, SamplingMessage):
    model: str
    stopReason: Literal["endTurn", "stopSequence", "maxTokens"] | str | None = None


# -----------------------------------------------------------------------------
# Completion (client -> server)
# -----------------------------------------------------------------------------


class ResourceTemplateReference(BaseModel):
    type: Literal["ref/resource"]
    uri: str


class PromptReference(BaseMetadata):
    type: Literal["ref/prompt"]


class CompletionArgument(BaseModel):
    name: str
    value: str


class CompletionContext(BaseModel):
    arguments: dict[str, str] | None = None


class CompleteParams(BaseModel):
    ref: ResourceTemplateReference | PromptReference
    argument: CompletionArgument
    context: CompletionContext | None = None


class CompleteRequest(JSONRPCRequest):
    method: Literal["completion/complete"] = Field(default="completion/complete", frozen=True)
    params: CompleteParams


class Completion(BaseModel):
    values: list[str]
    total: int | None = None
    hasMore: bool | None = None


class CompleteResult(Result):
    completion: Completion


# -----------------------------------------------------------------------------
# Roots
# -----------------------------------------------------------------------------


class ListRootsRequest(JSONRPCRequest):
    method: Literal["roots/list"] = Field(default="roots/list", frozen=True)


class Root(BaseModel):
    uri: str
    name: str | None = None
    meta: dict[str, Any] | None = Field(alias="_meta", default=None)


class ListRootsResult(Result):
    roots: list[Root]


class RootsListChangedNotification(JSONRPCMessage, Notification):
    method: Literal["notifications/roots/list_changed"] = Field(
        default="notifications/roots/list_changed", frozen=True
    )


# -----------------------------------------------------------------------------
# Elicitation (server -> client)
# -----------------------------------------------------------------------------

ElicitRequestedSchema = dict[str, Any]


class ElicitParams(BaseModel):
    message: str
    requestedSchema: ElicitRequestedSchema


class ElicitRequest(JSONRPCRequest):
    method: Literal["elicitation/create"] = Field(default="elicitation/create", frozen=True)
    params: ElicitParams


class ElicitResult(Result):
    action: Literal["accept", "decline", "cancel"]
    content: dict[str, str | int | float | bool | None] | None = None


# -----------------------------------------------------------------------------
# Union for middleware typing and convenience
# -----------------------------------------------------------------------------

MCPMessage = (
    JSONRPCRequest
    | JSONRPCResponse[Any]
    | JSONRPCError
    | InitializedNotification
    | CancelledNotification
    | ProgressNotification
    | LoggingMessageNotification
    | ResourceListChangedNotification
    | ResourceUpdatedNotification
    | PromptListChangedNotification
    | ToolListChangedNotification
    | RootsListChangedNotification
)



"""
MCP Context System

Provides the primary Context class for MCP tool development. This module contains
the Context class that tools should use for both runtime capabilities and
tool-specific data access.

The Context class combines:
- Runtime capabilities: logging, resources, prompts, sampling, UI, notifications
- Tool-specific data: secrets, user_id, authorization, metadata
- Session management: request/session IDs and MCP protocol handling

Key responsibilities:
- Manage per-request state and the current model context using a ContextVar
- Expose namespaced runtime capabilities (log, resources, etc.)
- Provide access to tool-specific data (secrets, user_id, etc.)
- Delegate to the underlying MCP session and server managers
- Handle MCP protocol communication and lifecycle management

Note: Context instances are automatically created and managed by the MCP server.
Tools receive a populated context instance as a parameter and should not
create Context instances directly.
"""

from __future__ import annotations

import asyncio
import logging
import uuid
import weakref
from builtins import list as builtins_list
from contextvars import ContextVar, Token
from typing import Any, cast

from arcade_core.context import ModelContext as ModelContextProtocol
from arcade_core.schema import (
    ToolContext,
)

from arcade_mcp_server.resource_server.base import ResourceOwner
from arcade_mcp_server.types import (
    AudioContent,
    CallToolParams,
    CallToolRequest,
    CallToolResult,
    ClientCapabilities,
    CreateMessageResult,
    ElicitResult,
    ImageContent,
    JSONRPCError,
    LoggingLevel,
    ModelHint,
    ModelPreferences,
    ResourceContents,
    Root,
    SamplingMessage,
    TextContent,
)

# Context variable for current model context
_current_model_context: ContextVar[Context | None] = ContextVar("model_context", default=None)
_flush_lock = asyncio.Lock()


class _ContextComponent:
    def __init__(self, ctx: Context) -> None:
        self._ctx = ctx

    @property
    def server(self) -> Any:
        return self._ctx.server

    def _require_session(self) -> Any:
        session = self._ctx._session
        if session is None:
            raise ValueError("Session not available")
        return session


class Context(ToolContext):
    """Primary context interface for MCP tools.

    This class provides both runtime capabilities (logging, resources, prompts, etc.)
    and tool-specific data (secrets, user_id, authorization) in a single interface.
    Tools should annotate their context parameter with this class.

    Runtime Capabilities:
        - log: Logging interface (context.log.info(), context.log.error(), etc.)
        - progress: Progress reporting for long-running operations
        - resources: Access to MCP resources (files, data sources, etc.)
        - tools: Call other tools programmatically
        - prompts: Access to MCP prompts and templates
        - sampling: Create messages using the client's model
        - ui: User interaction (elicit input from user)
        - notifications: Send notifications to the client

    Tool-Specific Data (inherited from ToolContext):
        - user_id: The user ID for this tool execution
        - secrets: List of secrets available to this tool
        - authorization: Authorization context if required
        - metadata: Additional metadata for the tool execution

    Example:
        ```python
        from arcade_mcp_server import Context, tool

        @tool
        async def my_tool(context: Context) -> str:
            '''Example tool'''
            # Runtime capabilities
            await context.log.info("Processing request")

            return "result"
        ```

    Note: Instances are automatically created and managed by the MCP server.
    Tools receive a populated context instance as a parameter.
    """

    # Mark as implementing the protocol
    __protocols__ = (ModelContextProtocol,) if ModelContextProtocol is not object else ()

    def __init__(
        self,
        server: Any,
        session: Any | None = None,
        request_id: str | None = None,
        resource_owner: ResourceOwner | None = None,
    ):
        """Initialize context with server reference."""
        super().__init__()
        self._server: weakref.ref[Any] = weakref.ref(server)
        self._session: Any | None = session
        self._tokens: list[Token] = []
        self._notification_queue: set[str] = set()
        self._request_id: str | None = request_id

        # Resource owner from front-door auth (if the server is protected)
        self._resource_owner: ResourceOwner | None = resource_owner

        # Namespaced adapters
        self._log = Logs(self)
        self._progress = Progress(self)
        self._resources = Resources(self)
        self._tools = Tools(self)
        self._prompts = Prompts(self)
        self._sampling = Sampling(self)
        self._ui = UI(self)
        self._notifications = Notifications(self)

    @property
    def server(self) -> Any:
        """Get the server instance."""
        server = self._server()
        if server is None:
            raise RuntimeError("Server instance is no longer available")
        return server

    def set_session(self, session: Any) -> None:
        """Set the session for this context."""
        self._session = session

    def set_request_id(self, request_id: str) -> None:
        """Set the request ID for this context."""
        self._request_id = request_id

    def set_tool_context(
        self,
        toolContext: ToolContext,
    ) -> None:
        """Populate the tool context fields for this model context."""
        self.authorization = toolContext.authorization
        self.secrets = toolContext.secrets
        self.metadata = toolContext.metadata
        self.user_id = toolContext.user_id

    async def __aenter__(self) -> Context:
        """Enter the context manager and set as current model context."""
        # Set this as current model context
        token = _current_model_context.set(self)
        self._tokens.append(token)
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """Exit the context manager and clear current model context."""
        # Flush any pending notifications
        await self._flush_notifications()

        # Reset context
        if self._tokens:
            token = self._tokens.pop()
            _current_model_context.reset(token)

    # ============ ModelContext protocol properties ============
    @property
    def log(self) -> Logs:
        """Logging interface for the tool.

        Provides methods for different log levels:
        - log.debug(message): Debug-level logging
        - log.info(message): Info-level logging
        - log.warning(message): Warning-level logging
        - log.error(message): Error-level logging
        - log.log(level, message): Log at a specific level

        Example:
            ```python
            await context.log.info("Processing started")
            await context.log.error("Something went wrong")
            ```
        """
        return self._log

    @property
    def progress(self) -> Progress:
        """Progress reporting for long-running operations.

        Use this to report progress back to the client during lengthy operations.

        Example:
            ```python
            await context.progress.report(0.5, total=1.0, message="Halfway done")
            ```
        """
        return self._progress

    @property
    def resources(self) -> Resources:
        """Interface for accessing MCP resources"""
        return self._resources

    @property
    def tools(self) -> Tools:
        """Interface for calling other tools programmatically.

        Allows tools to call other tools within the same session.

        Example:
            ```python
            result = await context.tools.call_raw("other_tool", {"param": "value"})
            ```
        """
        return self._tools

    @property
    def prompts(self) -> Prompts:
        """Interface for accessing MCP prompts and templates"""
        return self._prompts

    @property
    def sampling(self) -> Sampling:
        """Create messages using the client's model.

        Allows tools to generate text using the connected model.

        Example:
            ```python
            response = await context.sampling.create_message(
                "Summarize this text: " + text,
                temperature=0.7
            )
            ```
        """
        return self._sampling

    @property
    def ui(self) -> UI:
        """User interaction (elicitation) capabilities.

        Provides methods for interacting with the user, such as eliciting input.

        Example:
            ```python
            result = await context.ui.elicit(
                "Please provide your name",
                schema={"type": "object", "properties": {"name": {"type": "string"}}}
            )
            ```
        """
        return self._ui

    @property
    def notifications(self) -> Notifications:
        """
        Interface for sending notifications to the client
        such as tool list changes.

        Example:
            ```python
            await context.notifications.tools.list_changed()
            ```
        """
        return self._notifications

    @property
    def request_id(self) -> str | None:
        """Get the current request ID.

        Returns:
            The unique identifier for this MCP request, or None if not available.
        """
        return self._request_id

    @property
    def session_id(self) -> str | None:
        """Get the current session ID.

        Returns:
            The unique identifier for this MCP session, or None if not available.
        """
        if self._session is None:
            return None
        return getattr(self._session, "session_id", None)

    # Private helpers
    def _check_client_capability(self, capability: ClientCapabilities) -> bool:
        """Check if client has a capability."""
        if self._session is None:
            return False
        return cast(bool, self._session.check_client_capability(capability))

    def _parse_model_preferences(
        self, prefs: ModelPreferences | str | list[str] | None
    ) -> ModelPreferences | None:
        """Parse model preferences into standard format."""
        if prefs is None:
            return None
        elif isinstance(prefs, ModelPreferences):
            return prefs
        elif isinstance(prefs, str):
            return ModelPreferences(hints=[ModelHint(name=prefs)])
        elif isinstance(prefs, list):
            return ModelPreferences(hints=[ModelHint(name=h) for h in prefs])
        else:
            raise ValueError(f"Invalid model preferences type: {type(prefs)}")

    def _try_flush_notifications(self) -> None:
        """Try to flush notifications if in async context."""
        try:
            loop = asyncio.get_running_loop()
            if loop and not loop.is_running():
                return
            flush_task = asyncio.create_task(self._flush_notifications())
            flush_task.add_done_callback(lambda _: self._notification_queue.clear())
        except RuntimeError:
            # No event loop
            pass

    async def _flush_notifications(self) -> None:
        """Send all queued notifications."""
        async with _flush_lock:
            if not self._notification_queue or self._session is None:
                return

            nm = getattr(self.server, "notification_manager", None)
            if nm is None:
                return

            try:
                client_ids = []
                if (
                    self._session
                    and hasattr(self._session, "session_id")
                    and self._session.session_id
                ):
                    client_ids = [self._session.session_id]

                if "notifications/tools/list_changed" in self._notification_queue:
                    await nm.notify_tool_list_changed(client_ids)
                if "notifications/resources/list_changed" in self._notification_queue:
                    await nm.notify_resource_list_changed(client_ids)
                if "notifications/prompts/list_changed" in self._notification_queue:
                    pass

                self._notification_queue.clear()
            except Exception:
                # Don't let notification failures break the request
                logging.debug("Failed to send notifications", exc_info=True)


# =====================
# Namespaced adapters
# =====================
# These thin, per-domain facades (log, progress, resources, tools, prompts,
# sampling, ui, notifications) expose a stable, developer-friendly API on
# Context (e.g., context.log.info(...), context.resources.list()).
#
# They delegate all work to the active MCP session and server managers, keeping
# transport and server-specific details encapsulated in one place.
# This design:
# - avoids leaking MCP internals into the developer surface
# - preserves a cohesive, testable Context API with clear async boundaries
# - allows runtime implementations to evolve without breaking tool code
#
# In short: adapters provide the ergonomics tools rely on, while the underlying
# implementation remains decoupled and replaceable.


class Logs(_ContextComponent):
    def __init__(self, ctx: Context) -> None:
        super().__init__(ctx)

    async def log(
        self,
        level: str,
        message: str,
        logger_name: str | None = None,
        extra: dict[str, Any] | None = None,
    ) -> None:
        session = self._ctx._session
        if session is None:
            return
        level_typed = cast(LoggingLevel, level)
        data = {"msg": message, "extra": extra}
        await session.send_log_message(
            level=level_typed,
            data=data,
            logger=logger_name,
        )

    async def __call__(
        self,
        level: str,
        message: str,
        logger_name: str | None = None,
        extra: dict[str, Any] | None = None,
    ) -> None:  # compatibility shim
        await self.log(level, message, logger_name=logger_name, extra=extra)

    async def debug(self, message: str, **kwargs: Any) -> None:
        await self.log("debug", message, **kwargs)

    async def info(self, message: str, **kwargs: Any) -> None:
        await self.log("info", message, **kwargs)

    async def warning(self, message: str, **kwargs: Any) -> None:
        await self.log("warning", message, **kwargs)

    async def error(self, message: str, **kwargs: Any) -> None:
        await self.log("error", message, **kwargs)


class Progress(_ContextComponent):
    def __init__(self, ctx: Context) -> None:
        super().__init__(ctx)

    async def report(
        self, progress: float, total: float | None = None, message: str | None = None
    ) -> None:
        session = self._ctx._session
        if session is None:
            return
        progress_token = None
        if hasattr(session, "_request_meta") and session._request_meta is not None:
            progress_token = getattr(session._request_meta, "progressToken", None)
        if progress_token is None:
            return
        await session.send_progress_notification(
            progress_token=progress_token,
            progress=progress,
            total=total,
            message=message,
        )


class Resources(_ContextComponent):
    def __init__(self, ctx: Context) -> None:
        super().__init__(ctx)

    async def read(self, uri: str) -> list[ResourceContents]:
        if self._ctx.server is None:
            raise ValueError("Context is not available outside of a request")
        result = await self._ctx.server._mcp_read_resource(uri)
        return cast(list[ResourceContents], result)

    async def get(self, uri: str) -> ResourceContents:
        contents = await self.read(uri)
        if not contents:
            raise ValueError(f"Resource not found: {uri}")
        return contents[0]

    async def list_roots(self) -> list[Root]:
        if self._ctx._session is None:
            return []
        result = await self._ctx._session.list_roots()
        return result.roots if hasattr(result, "roots") else []

    async def list(self) -> list[Root]:
        # Convert Resource objects to Root objects
        resources = await self._ctx.server._resource_manager.list_resources()
        # Resources have uri and name which map to Root
        return [Root(uri=r.uri, name=r.name) for r in resources]

    async def list_templates(self) -> builtins_list[Any]:
        templates = await self._ctx.server._resource_manager.list_resource_templates()
        return cast(builtins_list[Any], templates)


class Tools(_ContextComponent):
    def __init__(self, ctx: Context) -> None:
        super().__init__(ctx)

    async def list(self) -> list[Any]:
        tools = await self._ctx.server._tool_manager.list_tools()
        return cast(list[Any], tools)

    async def call_raw(self, name: str, params: dict[str, Any]) -> CallToolResult:
        internal_id = f"internal-{uuid.uuid4()}"
        request = CallToolRequest(
            id=internal_id,
            params=CallToolParams(name=name, arguments=params),
        )

        response = await self._ctx.server._handle_call_tool(request, self._ctx._session)

        if isinstance(response, JSONRPCError):
            error_message = response.error.get("message", "Unknown error")
            return CallToolResult(
                content=[TextContent(type="text", text=error_message)],
                structuredContent={"error": error_message},
                isError=True,
            )

        return cast(CallToolResult, response.result)


class Prompts(_ContextComponent):
    def __init__(self, ctx: Context) -> None:
        super().__init__(ctx)

    async def list(self) -> list[Any]:
        prompts = await self._ctx.server._prompt_manager.list_prompts()
        return cast(list[Any], prompts)

    async def get(self, name: str, arguments: dict[str, str] | None = None) -> Any:
        return await self._ctx.server._prompt_manager.get_prompt(name, arguments)


class Sampling(_ContextComponent):
    def __init__(self, ctx: Context) -> None:
        super().__init__(ctx)

    async def create_message(
        self,
        messages: str | list[str | SamplingMessage],
        system_prompt: str | None = None,
        include_context: str | None = None,
        temperature: float | None = None,
        max_tokens: int | None = None,
        model_preferences: ModelPreferences | str | list[str] | None = None,
    ) -> TextContent | ImageContent | AudioContent | CreateMessageResult:
        if self._ctx._session is None:
            raise ValueError("Session not available")

        # Convert messages to proper format
        if isinstance(messages, str):
            sampling_messages = [
                SamplingMessage(content=TextContent(text=messages, type="text"), role="user")
            ]
        elif isinstance(messages, list):
            sampling_messages = []
            for m in messages:
                if isinstance(m, str):
                    sampling_messages.append(
                        SamplingMessage(content=TextContent(text=m, type="text"), role="user")
                    )
                else:
                    sampling_messages.append(m)
        else:
            sampling_messages = messages

        # Parse model preferences
        parsed_prefs = self._ctx._parse_model_preferences(model_preferences)

        # Check client capabilities
        if not self._ctx._check_client_capability(ClientCapabilities(sampling={})):
            raise ValueError("Client does not support sampling")

        result = await self._ctx._session.create_message(
            messages=sampling_messages,
            system_prompt=system_prompt,
            include_context=include_context,
            temperature=temperature,
            max_tokens=max_tokens or 512,
            model_preferences=parsed_prefs,
        )

        return result.content if hasattr(result, "content") else result  # type: ignore[no-any-return]


class UI(_ContextComponent):
    def __init__(self, ctx: Context) -> None:
        super().__init__(ctx)

    def _validate_elicitation_schema(self, schema: dict[str, Any]) -> None:
        """Validate that the schema conforms to MCP elicitation restrictions."""
        if not isinstance(schema, dict):
            raise TypeError("Schema must be a dictionary")

        if schema.get("type") != "object":
            raise ValueError("Schema must have type 'object'")

        properties = schema.get("properties", {})
        if not isinstance(properties, dict):
            raise TypeError("Schema properties must be a dictionary")

        # Validate each property
        for prop_name, prop_schema in properties.items():
            if not isinstance(prop_schema, dict):
                raise TypeError(f"Property '{prop_name}' schema must be a dictionary")

            prop_type = prop_schema.get("type")
            if prop_type not in ["string", "number", "integer", "boolean"]:
                raise ValueError(
                    f"Property '{prop_name}' has unsupported type '{prop_type}'. Only primitive types are allowed."
                )

            # Validate string formats
            if prop_type == "string" and "format" in prop_schema:
                allowed_formats = ["email", "uri", "date", "date-time"]
                if prop_schema["format"] not in allowed_formats:
                    raise ValueError(
                        f"Property '{prop_name}' has unsupported format '{prop_schema['format']}'. Allowed: {allowed_formats}"
                    )

    async def elicit(
        self, message: str, schema: dict[str, Any] | None = None, timeout: float = 300.0
    ) -> ElicitResult:
        if self._ctx._session is None:
            raise ValueError("Session not available")
        if schema is None:
            schema = {"type": "object", "properties": {}}

        # Validate schema conforms to MCP restrictions
        self._validate_elicitation_schema(schema)

        result = await self._ctx._session.elicit(
            message=message,
            requested_schema=schema,
            timeout=timeout,
        )
        return cast(ElicitResult, result)


class _NotificationsTools(_ContextComponent):
    def __init__(self, ctx: Context) -> None:
        super().__init__(ctx)

    async def list_changed(self) -> None:
        self._ctx._notification_queue.add("notifications/tools/list_changed")
        self._ctx._try_flush_notifications()


class _NotificationsResources(_ContextComponent):
    def __init__(self, ctx: Context) -> None:
        super().__init__(ctx)

    async def list_changed(self) -> None:
        self._ctx._notification_queue.add("notifications/resources/list_changed")
        self._ctx._try_flush_notifications()


class _NotificationsPrompts(_ContextComponent):
    def __init__(self, ctx: Context) -> None:
        super().__init__(ctx)

    async def list_changed(self) -> None:
        self._ctx._notification_queue.add("notifications/prompts/list_changed")
        self._ctx._try_flush_notifications()


class Notifications(_ContextComponent):
    def __init__(self, ctx: Context) -> None:
        super().__init__(ctx)
        self._tools = _NotificationsTools(ctx)
        self._resources = _NotificationsResources(ctx)
        self._prompts = _NotificationsPrompts(ctx)

    @property
    def tools(self) -> _NotificationsTools:
        return self._tools

    @property
    def resources(self) -> _NotificationsResources:
        return self._resources

    @property
    def prompts(self) -> _NotificationsPrompts:
        return self._prompts


def get_current_model_context() -> Context | None:
    """Get the current model context if available."""
    return _current_model_context.get()


def set_current_model_context(context: Context | None, token: Token | None = None) -> Token:
    """Set the current model context and return a token to reset it."""
    if token is not None:
        _current_model_context.reset(token)
        return token
    return _current_model_context.set(context)
